#![allow(dead_code)]

use std::{
    collections::{BTreeMap, BTreeSet},
    fmt::{Debug, Display},
    sync::Arc,
};

use crate::{
    buffers::{Buffer, BufferDesc},
    commands, def_monotonic_id,
    device::{self, DeviceOwned},
    images::{self, Image, ImageDesc},
    util::{self, Rgba, WithLifetime},
    SwapchainFrame,
};
use ash::vk;
use bytemuck::Contiguous;
use itertools::Itertools;
use petgraph::{
    graph::NodeIndex,
    visit::{EdgeRef, IntoNodeReferences, NodeRef},
};

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
#[repr(transparent)]
pub struct GraphResourceId(pub(crate) u32);

impl Display for GraphResourceId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "#{}", self.0)
    }
}

#[derive(Debug, Clone)]
pub enum GraphResourceDesc {
    Image(ImageDesc),
    Buffer(BufferDesc),
}

impl From<GraphResourceDesc> for GraphResource {
    fn from(value: GraphResourceDesc) -> Self {
        match value {
            GraphResourceDesc::Image(image_desc) => Self::ImageDesc(image_desc),
            GraphResourceDesc::Buffer(buffer_desc) => Self::BufferDesc(buffer_desc),
        }
    }
}

#[derive(Default, Debug, PartialEq, Eq)]
pub enum GraphResource {
    Framebuffer(Arc<SwapchainFrame>),
    ImportedImage(Arc<Image>),
    ImportedBuffer(Arc<Buffer>),
    Image(Arc<Image>),
    Buffer(Buffer),
    ImageDesc(ImageDesc),
    BufferDesc(BufferDesc),
    #[default]
    Default,
}

impl GraphResource {
    fn simple_hash(&self) -> u64 {
        use std::hash::{Hash, Hasher};
        let mut state = std::hash::DefaultHasher::new();
        let discr = core::mem::discriminant(self);
        discr.hash(&mut state);

        match self {
            GraphResource::Framebuffer(swapchain_frame) => {
                (swapchain_frame.index, swapchain_frame.image.handle()).hash(&mut state)
            }
            GraphResource::ImportedImage(image) => image.handle().hash(&mut state),
            GraphResource::ImportedBuffer(buffer) => buffer.handle().hash(&mut state),
            GraphResource::Image(image) => image.handle().hash(&mut state),
            GraphResource::Buffer(buffer) => buffer.handle().hash(&mut state),
            GraphResource::ImageDesc(image_desc) => image_desc.hash(&mut state),
            GraphResource::BufferDesc(buffer_desc) => buffer_desc.hash(&mut state),
            GraphResource::Default => {}
        }

        state.finish()
    }
}

#[derive(Debug, Clone, Copy)]
pub enum LoadOp {
    Clear(Rgba),
    Load,
    DontCare,
}

#[derive(Debug, Clone, Copy)]
pub enum StoreOp {
    DontCare,
    Store,
}

pub struct RenderContext<'a> {
    pub device: device::Device,
    pub cmd: commands::SingleUseCommand,
    pub resources: &'a [GraphResource],
}

impl RenderContext<'_> {
    pub fn get_image(&self, id: GraphResourceId) -> Option<&Arc<Image>> {
        self.resources.get(id.0 as usize).and_then(|res| match res {
            GraphResource::ImportedImage(arc) => Some(arc),
            GraphResource::Image(image) => Some(image),
            GraphResource::Framebuffer(fb) => Some(&fb.image),
            _ => None,
        })
    }
    pub fn get_buffer(&self, id: GraphResourceId) -> Option<&Buffer> {
        self.resources.get(id.0 as usize).and_then(|res| match res {
            GraphResource::ImportedBuffer(arc) => Some(arc.as_ref()),
            GraphResource::Buffer(buffer) => Some(buffer),
            _ => None,
        })
    }
}

#[derive(Debug, Clone, Copy, Default, PartialEq, Eq, PartialOrd, Ord)]
pub struct Access {
    pub stage: vk::PipelineStageFlags2,
    pub mask: vk::AccessFlags2,
    pub layout: Option<vk::ImageLayout>,
}

impl core::ops::BitOr for Access {
    type Output = Self;

    fn bitor(self, rhs: Self) -> Self::Output {
        //assert_eq!(self.layout, rhs.layout);
        Self {
            stage: self.stage | rhs.stage,
            mask: self.mask | rhs.mask,
            layout: self.layout.max(rhs.layout),
        }
    }
}

#[derive(Debug, Clone, Copy, Default, PartialEq, Eq, PartialOrd, Ord)]
pub struct AccessMask {
    pub stage: vk::PipelineStageFlags2,
    pub mask: vk::AccessFlags2,
}
impl AccessMask {
    pub fn empty() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
        }
    }
    pub fn is_empty(&self) -> bool {
        self.stage.is_empty() && self.mask.is_empty()
    }
}
impl core::ops::BitOr for AccessMask {
    type Output = Self;

    fn bitor(self, rhs: Self) -> Self::Output {
        Self {
            stage: self.stage | rhs.stage,
            mask: self.mask | rhs.mask,
        }
    }
}
impl core::ops::Not for AccessMask {
    type Output = Self;

    fn not(self) -> Self::Output {
        Self {
            stage: !self.stage,
            mask: !self.mask,
        }
    }
}

impl core::ops::BitAnd for AccessMask {
    type Output = Self;

    fn bitand(self, rhs: Self) -> Self::Output {
        Self {
            stage: self.stage & rhs.stage,
            mask: self.mask & rhs.mask,
        }
    }
}

impl core::ops::BitXor for AccessMask {
    type Output = Self;

    fn bitxor(self, rhs: Self) -> Self::Output {
        Self {
            stage: self.stage ^ rhs.stage,
            mask: self.mask ^ rhs.mask,
        }
    }
}

impl Access {
    pub fn into_access_mask(&self) -> AccessMask {
        AccessMask {
            stage: self.stage,
            mask: self.mask,
        }
    }
    pub fn empty() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: None,
        }
    }
    pub fn undefined() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::UNDEFINED),
        }
    }
    /// Only use this for `Ord`!
    pub fn min() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::from_raw(u64::MAX),
            mask: vk::AccessFlags2::from_raw(u64::MAX),
            layout: None,
        }
    }
    /// Only use this for `Ord`!
    pub fn max() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::from_raw(u64::MAX),
            mask: vk::AccessFlags2::from_raw(u64::MAX),
            layout: Some(vk::ImageLayout::from_raw(i32::MAX)),
        }
    }
    pub fn general() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::GENERAL),
        }
    }
    pub fn transfer_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::TRANSFER,
            mask: vk::AccessFlags2::TRANSFER_READ,
            layout: Some(vk::ImageLayout::TRANSFER_SRC_OPTIMAL),
        }
    }
    pub fn transfer_write() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::TRANSFER,
            mask: vk::AccessFlags2::TRANSFER_WRITE,
            layout: Some(vk::ImageLayout::TRANSFER_DST_OPTIMAL),
        }
    }
    pub fn vertex_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::VERTEX_ATTRIBUTE_INPUT,
            mask: vk::AccessFlags2::VERTEX_ATTRIBUTE_READ,
            layout: None,
        }
    }
    pub fn index_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::INDEX_INPUT,
            mask: vk::AccessFlags2::INDEX_READ,
            layout: None,
        }
    }
    pub fn indirect_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::DRAW_INDIRECT,
            mask: vk::AccessFlags2::INDIRECT_COMMAND_READ,
            layout: None,
        }
    }
    pub fn color_attachment_read_only() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::COLOR_ATTACHMENT_OUTPUT,
            mask: vk::AccessFlags2::COLOR_ATTACHMENT_READ,
            layout: Some(vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL),
        }
    }
    pub fn color_attachment_write_only() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::COLOR_ATTACHMENT_OUTPUT,
            mask: vk::AccessFlags2::COLOR_ATTACHMENT_WRITE,
            layout: Some(vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL),
        }
    }
    pub fn color_attachment_read_write() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::COLOR_ATTACHMENT_OUTPUT,
            mask: vk::AccessFlags2::COLOR_ATTACHMENT_WRITE
                | vk::AccessFlags2::COLOR_ATTACHMENT_READ,
            layout: Some(vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL),
        }
    }
    pub fn present() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::PRESENT_SRC_KHR),
        }
    }
}

pub type RecordFn = dyn FnOnce(&RenderContext) -> crate::Result<()> + Send;

pub struct PassDesc {
    // this pass performs `Access` read on `GraphResourceId`.
    // some `GraphResourceId` may occur multiple times.
    pub reads: Vec<(GraphResourceId, Access)>,
    // this pass performs `Access` write on `GraphResourceId`.
    // some `GraphResourceId` may occur multiple times.
    pub writes: Vec<(GraphResourceId, Access)>,
    pub record: Box<RecordFn>,
}

impl Default for PassDesc {
    fn default() -> Self {
        Self {
            reads: Default::default(),
            writes: Default::default(),
            record: Box::new(|_| Ok(())),
        }
    }
}

impl Debug for PassDesc {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("PassDesc")
            .field("reads", &self.reads)
            .field("write", &self.writes)
            .finish_non_exhaustive()
    }
}

def_monotonic_id!(pub RenderGraphPassId);

// Non-imported resources remain `RenderGraphResourceDesc`s because they may be
// able to be aliased.
// This should be dual to liveness/register allocation in a compiler.
// Dummy-impl is just allocating every resource_desc itself. 5head-impl is trying
// to find resource_descs which are eq, but whose liveness doesn't overlap.
#[derive(Debug)]
pub struct RenderGraph {
    resources: Vec<GraphResource>,
    accesses: Vec<Access>,
    pass_descs: Vec<PassDesc>,
    /// the rendergraph produces these resources. Any passes on which these
    /// outputs do not depend are pruned.
    outputs: Vec<GraphResourceId>,
}

impl RenderGraph {
    pub fn new() -> Self {
        Self {
            resources: Vec::new(),
            pass_descs: Vec::new(),
            accesses: Vec::new(),
            outputs: Vec::new(),
        }
    }

    fn get_next_resource_id(&mut self) -> GraphResourceId {
        GraphResourceId(self.resources.len() as u32)
    }

    pub fn add_resource(&mut self, desc: GraphResourceDesc) -> GraphResourceId {
        let id = self.get_next_resource_id();
        self.resources.push(desc.into());
        self.accesses.push(Access::undefined());
        id
    }
    pub fn mark_as_output(&mut self, id: GraphResourceId) {
        // TODO: dedup
        self.outputs.push(id);
    }
    pub fn import_resource(&mut self, res: GraphResource, access: Access) -> GraphResourceId {
        if let Some(i) = self
            .resources
            .iter()
            .position(|other| res.simple_hash() == other.simple_hash())
        {
            GraphResourceId(i as u32)
        } else {
            let id = self.get_next_resource_id();
            self.resources.push(res);
            self.accesses.push(access);
            id
        }
    }
    pub fn import_image(&mut self, image: Arc<Image>, access: Access) -> GraphResourceId {
        let res = GraphResource::ImportedImage(image);
        self.import_resource(res, access)
    }
    pub fn import_buffer(&mut self, buffer: Arc<Buffer>, access: Access) -> GraphResourceId {
        let res = GraphResource::ImportedBuffer(buffer);
        self.import_resource(res, access)
    }
    pub fn import_framebuffer(&mut self, frame: Arc<SwapchainFrame>) -> GraphResourceId {
        self.import_resource(GraphResource::Framebuffer(frame), Access::undefined())
    }
    pub fn add_pass(&mut self, pass: PassDesc) {
        self.pass_descs.push(pass);
    }

    // https://blog.traverseresearch.nl/render-graph-101-f42646255636
    // https://github.com/EmbarkStudios/kajiya/blob/main/crates/lib/kajiya-rg/src/graph.rs
    // https://themaister.net/blog/2017/08/15/render-graphs-and-vulkan-a-deep-dive/
    pub fn resolve(
        &mut self,
        device: device::Device,
    ) -> crate::Result<WithLifetime<'_, commands::CommandList<commands::SingleUseCommand>>> {
        let now = std::time::Instant::now();

        let mut refmap = util::asdf::NodeRefsMap::new(self.resources.len(), self.pass_descs.len());

        refmap.allocate_ref_ranges(&self.pass_descs);
        refmap.ref_passes(&self.pass_descs);
        refmap.ref_inputs(&self.resources);
        refmap.ref_outputs(&self.outputs);
        let dag = refmap.build_dag();
        let topo = refmap.toposort_dag(dag);

        eprintln!(
            "resolved render graph in {}ms",
            now.elapsed().as_micros() as f32 / 1e3
        );

        // create internal resources:
        for (i, res) in self.resources.iter_mut().enumerate() {
            match res {
                GraphResource::ImageDesc(image_desc) => {
                    tracing::trace!("creating resource #{i:?} with {image_desc:?}");
                    *res = GraphResource::Image(Arc::new(Image::new(
                        device.clone(),
                        image_desc.clone(),
                    )?));
                }
                GraphResource::BufferDesc(buffer_desc) => {
                    tracing::trace!("creating resource #{i:?} with {buffer_desc:?}");
                    *res = GraphResource::Buffer(Buffer::new(device.clone(), buffer_desc.clone())?);
                }
                _ => {}
            }
        }

        let now = std::time::Instant::now();
        let pool =
            commands::SingleUseCommandPool::new(device.clone(), device.graphics_queue().clone())?;

        let resources = &self.resources;
        let cmds = topo
            .into_iter()
            .rev()
            .map(|(passes, accesses)| {
                let passes = passes
                    .into_iter()
                    .filter_map(|i| i.unpack().get_pass_idx())
                    .map(|i| core::mem::take(&mut self.pass_descs[i]))
                    .collect::<Vec<_>>();
                (passes, accesses)
            })
            .map({
                |(passes, accesses)| {
                    let cmd = pool.alloc()?;
                    // transitions
                    for (&id, &(from, to)) in accesses.iter() {
                        Self::transition_resource(
                            &resources[id.0 as usize],
                            device.dev(),
                            unsafe { &cmd.buffer() },
                            from,
                            to,
                        );
                    }

                    let ctx = RenderContext {
                        device: device.clone(),
                        cmd,
                        resources,
                    };

                    for pass in passes {
                        (pass.record)(&ctx)?;
                    }

                    ctx.cmd.end()?;
                    crate::Result::Ok(ctx.cmd)
                }
            })
            .collect::<crate::Result<Vec<_>>>()?;

        let cmd_list = commands::CommandList(cmds);

        tracing::debug!(
            "recording render graph: {}ms",
            now.elapsed().as_micros() as f32 / 1e3
        );

        Ok(WithLifetime::new(cmd_list))
    }

    pub fn get_outputs(&mut self) -> BTreeMap<GraphResourceId, GraphResource> {
        let outputs = self
            .outputs
            .iter()
            .map(|id| (*id, core::mem::take(&mut self.resources[id.0 as usize])))
            .collect::<BTreeMap<_, _>>();

        outputs
    }

    pub fn transition_resource(
        res: &GraphResource,
        dev: &ash::Device,
        cmd: &vk::CommandBuffer,
        from: Access,
        to: Access,
    ) {
        let barrier: Barrier = match res {
            GraphResource::Framebuffer(arc) => {
                image_barrier(arc.image.handle(), arc.image.format(), from, to, None).into()
            }
            GraphResource::ImportedImage(arc) => {
                image_barrier(arc.handle(), arc.format(), from, to, None).into()
            }
            GraphResource::ImportedBuffer(arc) => {
                buffer_barrier(arc.handle(), 0, arc.len(), from, to, None).into()
            }
            GraphResource::Image(image) => {
                image_barrier(image.handle(), image.format(), from, to, None).into()
            }
            GraphResource::Buffer(buffer) => {
                buffer_barrier(buffer.handle(), 0, buffer.len(), from, to, None).into()
            }
            _ => {
                unreachable!()
            }
        };

        unsafe {
            dev.cmd_pipeline_barrier2(*cmd, &((&barrier).into()));
        }
    }

    fn transition_resource_to(
        accesses: &mut BTreeMap<GraphResourceId, Access>,
        resources: &BTreeMap<GraphResourceId, GraphResource>,
        dev: &ash::Device,
        cmd: &vk::CommandBuffer,
        id: GraphResourceId,
        to: Access,
    ) {
        let old_access = accesses.get(&id);
        let res = resources.get(&id);

        if let (Some(&old_access), Some(res)) = (old_access, res) {
            Self::transition_resource(res, dev, cmd, old_access, to);
            accesses.insert(id, to);
        }
    }
}

pub enum Barrier {
    Image(vk::ImageMemoryBarrier2<'static>),
    Buffer(vk::BufferMemoryBarrier2<'static>),
}

impl<'a> From<&'a Barrier> for vk::DependencyInfo<'a> {
    fn from(value: &'a Barrier) -> Self {
        let info = vk::DependencyInfo::default();
        let info = match value {
            Barrier::Image(barrier) => info.image_memory_barriers(core::slice::from_ref(barrier)),
            Barrier::Buffer(barrier) => info.buffer_memory_barriers(core::slice::from_ref(barrier)),
        };

        info
    }
}

impl From<vk::ImageMemoryBarrier2<'static>> for Barrier {
    fn from(value: vk::ImageMemoryBarrier2<'static>) -> Self {
        Self::Image(value)
    }
}
impl From<vk::BufferMemoryBarrier2<'static>> for Barrier {
    fn from(value: vk::BufferMemoryBarrier2<'static>) -> Self {
        Self::Buffer(value)
    }
}

pub fn buffer_barrier(
    buffer: vk::Buffer,
    offset: u64,
    size: u64,
    before: Access,
    after: Access,
    queue_families: Option<(u32, u32)>,
) -> vk::BufferMemoryBarrier2<'static> {
    vk::BufferMemoryBarrier2::default()
        .buffer(buffer)
        .offset(offset)
        .size(size)
        .src_access_mask(before.mask)
        .src_stage_mask(before.stage)
        .dst_access_mask(after.mask)
        .dst_stage_mask(after.stage)
        .src_queue_family_index(
            queue_families
                .map(|(src, _)| src)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
        .dst_queue_family_index(
            queue_families
                .map(|(_, dst)| dst)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
}

pub fn image_barrier(
    image: vk::Image,
    format: vk::Format,
    before_access: Access,
    after_access: Access,
    queue_families: Option<(u32, u32)>,
) -> vk::ImageMemoryBarrier2<'static> {
    vk::ImageMemoryBarrier2::default()
        .src_access_mask(before_access.mask)
        .src_stage_mask(before_access.stage)
        .dst_access_mask(after_access.mask)
        .dst_stage_mask(after_access.stage)
        .image(image)
        .old_layout(before_access.layout.unwrap_or_default())
        .new_layout(after_access.layout.unwrap_or_default())
        .subresource_range(vk::ImageSubresourceRange {
            aspect_mask: util::image_aspect_from_format(format),
            ..images::SUBRESOURCERANGE_ALL
        })
        .src_queue_family_index(
            queue_families
                .map(|(src, _)| src)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
        .dst_queue_family_index(
            queue_families
                .map(|(_, dst)| dst)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
}

// #[cfg(test)]
// mod tests {
//     use super::*;

//     macro_rules! def_dummy_pass {
//         ($name:ident: {$queue:path, $layout_in:path, $layout_out:path}) => {
//             #[derive(Debug, Clone)]
//             struct $name(Vec<RenderGraphResourceId>, Vec<RenderGraphResourceId>);
//             impl Pass for $name {
//                 fn get_read_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
//                     ResourceAccess::default()
//                 }
//                 fn get_write_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
//                     ResourceAccess::default()
//                 }

//                 fn get_queue_capability_requirements(&self) -> device::QueueFlags {
//                     $queue
//                 }

//                 fn get_read_dependencies<'a>(
//                     &'a self,
//                 ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
//                     Box::new(self.0.iter().cloned())
//                 }

//                 fn get_write_dependencies<'a>(
//                     &'a self,
//                 ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
//                     Box::new(self.1.iter().cloned())
//                 }

//                 fn record(&self, _ctx: &RenderContext) -> crate::Result<()> {
//                     Ok(())
//                 }
//             }
//         };
//     }

//     def_dummy_pass!(DepthPass: {
//         device::QueueFlags::GRAPHICS,
//         vk::ImageLayout::DEPTH_ATTACHMENT_OPTIMAL,
//         vk::ImageLayout::DEPTH_ATTACHMENT_OPTIMAL});
//     def_dummy_pass!(RenderPass: {
//         device::QueueFlags::GRAPHICS,
//         vk::ImageLayout::COLORiATTACHMENT_OPTIMAL,
//         vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL});
//     def_dummy_pass!(AsyncPass: {
//         device::QueueFlags::ASYNC_COMPUTE,
//         vk::ImageLayout::UNDEFINED,
//         vk::ImageLayout::GENERAL});
//     def_dummy_pass!(PostProcessPass: {
//         device::QueueFlags::ASYNC_COMPUTE,
//         vk::ImageLayout::GENERAL,
//         vk::ImageLayout::GENERAL});
//     def_dummy_pass!(PresentPass: {
//         device::QueueFlags::PRESENT,
//         vk::ImageLayout::PRESENT_SRC_KHR,
//         vk::ImageLayout::UNDEFINED});
//     def_dummy_pass!(DepthVisualisationPass: {
//         device::QueueFlags::ASYNC_COMPUTE,
//         vk::ImageLayout::GENERAL,
//         vk::ImageLayout::UNDEFINED});

//     #[test]
//     fn resolve_graph() {
//         let mut graph = RenderGraph::new();
//         let gbuffer = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
//             ..Default::default()
//         }));
//         let depth_image = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
//             ..Default::default()
//         }));
//         let depth_visualisation = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
//             ..Default::default()
//         }));
//         let compute_buffer = graph.add_resource(RenderGraphResourceDesc::Buffer(BufferDesc {
//             ..Default::default()
//         }));

//         graph.add_pass(DepthPass(vec![depth_image], vec![depth_image]));
//         graph.add_pass(DepthVisualisationPass(
//             vec![depth_image, depth_visualisation],
//             vec![depth_visualisation],
//         ));
//         graph.add_pass(AsyncPass(vec![compute_buffer], vec![compute_buffer]));
//         graph.add_pass(RenderPass(
//             vec![depth_image, compute_buffer, gbuffer],
//             vec![gbuffer],
//         ));
//         graph.add_pass(PostProcessPass(vec![gbuffer], vec![gbuffer]));
//         graph.mark_as_output(gbuffer);
//         graph.mark_as_output(depth_image);

//         // graph.resolve();
//     }
// }

pub fn clear_pass(rg: &mut RenderGraph, color: Rgba, target: GraphResourceId) {
    let reads = [(target, Access::transfer_write())].to_vec();
    let writes = [(target, Access::transfer_write())].to_vec();

    let record: Box<RecordFn> = Box::new({
        move |ctx| {
            let target = ctx.get_image(target).unwrap();
            let cmd = &ctx.cmd;

            cmd.clear_color_image(
                target.handle(),
                target.format(),
                vk::ImageLayout::TRANSFER_DST_OPTIMAL,
                color,
                &[images::SUBRESOURCERANGE_COLOR_ALL],
            );
            Ok(())
        }
    });
    rg.add_pass(PassDesc {
        reads,
        writes,
        record,
    });
}

pub fn present_pass(rg: &mut RenderGraph, target: GraphResourceId) {
    let record: Box<RecordFn> = Box::new(|_| Ok(()));
    let reads = vec![(target, Access::present())];
    let writes = vec![(target, Access::present())];
    rg.add_pass(PassDesc {
        reads,
        writes,
        record,
    });
}
